[<Microsoft.FSharp.Core.AutoOpen>]
module Degenz.Types

open System
open Degenz

[<Measure>]
type mins

[<Measure>]
type GBT

type Range = { Min : int ; Max : int }

module Range =
    let normalized = { Min = 0 ; Max = 100 }
    let create min max = { Min = min ; Max = max }
    let constrain value range = if value < range.Min then range.Min elif value > range.Max then range.Max else value

type ItemId =
    | Virus = 0
    | RemoteAccess = 1
    | Worm = 2
    | Firewall = 6
    | Encryption = 7
    | Cypher = 8
    | ProteinPowder = 12
    | ToroLoco = 13
    | Cigs = 14
    | MoonPie = 15

type StatId =
    | Strength = 0
    | Focus = 1
    | Charisma = 2
    | Luck = 3

type StatConfig = {
    Id : StatId
    BaseDecayRate : float
    BaseRange : Range
}

type PlayerStat = {
    Id : StatId
    Amount : int
    ModRange : Range
    LastRead : DateTime
}
with static member empty = { Id = StatId.Strength ; Amount = 0 ; ModRange = Range.normalized ; LastRead = DateTime.UtcNow}

type Stats = {
    Strength : PlayerStat
    Focus : PlayerStat
    Luck : PlayerStat
    Charisma : PlayerStat
}
with static member empty = { Strength = PlayerStat.empty ; Focus = PlayerStat.empty ; Luck = PlayerStat.empty ; Charisma = PlayerStat.empty }

type HackResult =
    | Strong
    | Weak

type DiscordPlayer = { Id: uint64; Name: string }
    with static member empty = { Id = 0uL ; Name = "None" }

type HackEvent = {
    IsInstigator : bool
    Adversary : DiscordPlayer
    Success : bool
    HackId : int
}

type PlayerEventType =
    | Hacking of HackEvent
    | Shielding of shieldId : int
    | Stealing of instigator : bool * adversary : DiscordPlayer
    | Imprison

type PlayerEvent =
    { Type : PlayerEventType
      Cooldown : int<mins>
      Timestamp : DateTime }

[<RequireQualifiedAccess>]
type ItemType =
    | Hack
    | Shield
    | Food
    | Accessory

type ItemAttributes = {
    CanBuy : bool
    CanSell : bool
    CanConsume : bool
    CanTrade : bool
    CanDrop : bool
}

type Item = {
    Id : int
    Name : string
    Price : int<GBT>
    Attributes : ItemAttributes
}

type HackItem = {
    Power : int
    Class : int
    Cooldown : int<mins>
    Item : Item
}

type ShieldItem = {
    Class : int
    Cooldown : int<mins>
    Item : Item
}

type FoodItem = {
    TargetStat : StatId
    BoostAmount : int
    Item : Item
}

type AccessoryItem = {
    TargetStat : StatId
    FloorBoost : int
    CeilBoost : int
    Item : Item
}

type ItemDetails =
    | Hack of HackItem
    | Shield of ShieldItem
    | Food of FoodItem
    | Accessory of AccessoryItem
    member this.Id =
        match this with
        | Hack i -> i.Item.Id
        | Shield i -> i.Item.Id
        | Food i -> i.Item.Id
        | Accessory i -> i.Item.Id
    member this.Name =
        match this with
        | Hack i -> i.Item.Name
        | Shield i -> i.Item.Name
        | Food i -> i.Item.Name
        | Accessory i -> i.Item.Name
    member this.Price =
        match this with
        | Hack i -> i.Item.Price
        | Shield i -> i.Item.Price
        | Food i -> i.Item.Price
        | Accessory i -> i.Item.Price
    member this.getItem =
        match this with
        | Hack i -> i.Item
        | Shield i -> i.Item
        | Food i -> i.Item
        | Accessory i -> i.Item
    static member getClass = function
        | Hack i -> i.Class
        | Shield i -> i.Class
        | Food _ -> -1
        | Accessory _ -> -1

type Inventory = ItemDetails list

type PlayerData = {
    DiscordId : uint64
    Name : string
    Inventory : Inventory
    Events : PlayerEvent list
    Stats : Stats
    Bank : int<GBT>
}
//        Achievements : string array
//        XP : int
with member this.toDiscordPlayer = { Id = this.DiscordId ; Name = this.Name }
     static member empty =
         { DiscordId = 0uL
           Name = "None"
           Inventory = []
           Events = []
           Stats = Stats.empty
//               Achievements = [||]
//               XP = 0
           Bank = 0<GBT> }
